enum EventName {
  RefreshToken = 'refreshToken'
}

/**
 * Dispatch a event to App. Web->App
 */
const send = (
  eventName: Anue.Utils.NativeEventName,
  payload: Record<string, any> = {}
) => {
  if (typeof native !== 'undefined') {
    try {
      switch (eventName) {
        case 'closeWebView':
          native.member_event(JSON.stringify({ action: 'anueCloseWebView' }));
          break;
        case 'anueRequestRefreshToken':
          native.member_event(
            JSON.stringify({ action: 'anueRequestRefreshToken' })
          );
          break;
        case 'anueMemberLevelChanged':
          native.member_event(
            JSON.stringify({ action: 'anueMemberLevelChanged' })
          );
          break;
        case 'anueAlert':
          native.member_event(
            JSON.stringify({
              action: 'anueAlert',
              title: payload.title || 'anueAlert is called',
              message: payload.message || 'anueAlert is called'
            })
          );
        default:
        // TODO: Move the implementation to SDK from fe-portal.
        // native.event(JSON.stringify(payload));
      }
    } catch (e) {
      console.warn(`%cError occurs during communicate with APP`, e);
    }
  }
};

/**
 * Init the listener for App. App->Web
 */
const addListener = {
  webViewStatus() {
    if (typeof window !== 'undefined') {
      window.appCallWebCurrentStatus = (status: number) => {
        switch (status) {
          case 0:
            window.webViewStatus = 'onResume';
            break;
          case 1:
            window.webViewStatus = 'onPause';
            break;
          default:
            window.webViewStatus = status;
        }

        return 'appCallWebCurrentStatus is called from App';
      };
    }
  },
  webViewRefreshToken() {
    if (typeof window !== 'undefined') {
      console.log('%c[WebView] appCallWebRefreshToken', 'color: red;');
      window.appCallWebRefreshToken = (token: string) => {
        const prevToken = window.webViewToken;

        console.log('token received', token);
        if (prevToken !== token) {
          const event = new CustomEvent(EventName.RefreshToken, {
            detail: token
          });

          window.webViewToken = token;
          window.dispatchEvent(event);
        }

        return 'appCallWebRefreshToken is called from App';
      };
    }
  }
};

interface AnueNativeEvent extends Event {
  detail: string;
}

const get = {
  webViewStatus: () => window.webViewStatus,
  async webViewRefreshToken() {
    if (typeof window !== 'undefined') {
      return new Promise((resolve, reject) => {
        function handleRefreshTokenResponse(event: AnueNativeEvent) {
          const token = event.detail;

          window.removeEventListener(
            EventName.RefreshToken,
            handleRefreshTokenResponse as EventListener
          );

          if (token) {
            console.log('Receive Token from App');
            send('anueAlert', { title: 'anueAlert - Received token' });
            resolve(token);
          } else {
            reject(new Error('Refresh token failed in the webview'));
            send('anueAlert', { title: 'anueAlert - Refresh token timeout!' });
          }
        }

        setTimeout(() => {
          alert('Failed');
          reject(new Error('Refresh token time out!'));
        }, 5000);

        window.addEventListener(
          EventName.RefreshToken,
          handleRefreshTokenResponse as EventListener
        );
        send('anueRequestRefreshToken');
      });
    }
  }
};

const webview = {
  send,
  addListener,
  get
};

anue.shared.set('utils.webview', webview);

export default webview;
